"""Entry point for ``python -m byteb4rb1e.homeostat``."""

from byteb4rb1e.utils.argparse.dispatcher import CLI

from byteb4rb1e.homeostat.cli.commands import (
    AuthorizeCommand,
    DrinkCommand,
    MeasureCommand,
    MigrateCommand,
    ProjectCommand,
    SyncCommand,
)


def main() -> None:
    cli = CLI(
        prog="homeostat",
        description="Health data ETL pipeline",
    )
    cli.bootstrap([
        AuthorizeCommand,
        SyncCommand,
        ProjectCommand,
        MeasureCommand,
        DrinkCommand,
        MigrateCommand,
    ])
    cli.run()


if __name__ == "__main__":
    main()
